use anyhow::Error;
use gc_arena::{Gc, Rootable};
use piccolo::{self, Callback, Context, FromValue, Function, Table, UserData, Value};
use wasm_bindgen::JsValue;
use web_sys::console;
use yew::UseStateSetter;

use crate::{
    command_handler::debrace,
    id_intern::{intern_id, unintern_id},
    websocket::{connect_websocket, send_message_to_mud, WebSocketId},
    GlobalLayoutCell, GlobalMemoCell,
};

use super::LuaState;

fn try_unwrap_socketid<'gc>(
    ctx: Context<'gc>,
    value: &Value<'gc>,
) -> Result<WebSocketId, piccolo::Error<'gc>> {
    let value = if let Ok(sockname) = String::from_value(ctx, *value) {
        let ret = Table::from_value(ctx, ctx.get_global("muds"))?.get(ctx, sockname);
        if ret.is_nil() {
            Err(Error::msg(
                "Could not find a MUD connection with that name.",
            ))?;
        }
        ret
    } else {
        *value
    };
    Ok(UserData::from_value(ctx, value)?
        .downcast::<Rootable!['gcb => Gc<'gcb, WebSocketId>]>()?
        .as_ref()
        .clone())
}

pub fn handle_websocket_output(
    socket: &WebSocketId,
    engine: &mut LuaState,
    input: &[u8],
) -> Result<(), String> {
    engine
        .interp
        .try_enter(|ctx| {
            let handlers = Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"handlers")))?;
            let input_fn =
                Function::from_value(ctx, handlers.get(ctx, ctx.intern_static(b"mudoutput")))?;
            ctx.fetch(&engine.exec).restart(
                ctx,
                input_fn,
                (
                    intern_id::<WebSocketId>(ctx, socket.clone()),
                    ctx.intern(input),
                ),
            );
            Ok(())
        })
        .map_err(|err| format!("{}", err))?;
    engine
        .interp
        .execute::<()>(&engine.exec)
        .map_err(|err| format!("{}", err))
}

pub fn handle_websocket_output_log_err(socket: &WebSocketId, engine: &mut LuaState, input: &[u8]) {
    match handle_websocket_output(socket, engine, input) {
        Ok(()) => {}
        Err(e) => console::log_2(
            &JsValue::from_str("An error occurred calling the WebSocket input handler"),
            &JsValue::from_str(&e),
        ),
    }
}

pub fn handle_websocket_close(socket: &WebSocketId, engine: &mut LuaState) -> Result<(), String> {
    engine
        .interp
        .try_enter(|ctx| {
            let handlers = Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"handlers")))?;
            let input_fn =
                Function::from_value(ctx, handlers.get(ctx, ctx.intern_static(b"mudclose")))?;
            ctx.fetch(&engine.exec).restart(
                ctx,
                input_fn,
                intern_id::<WebSocketId>(ctx, socket.clone()),
            );
            Ok(())
        })
        .map_err(|err| format!("{}", err))?;
    engine
        .interp
        .execute::<()>(&engine.exec)
        .map_err(|err| format!("{}", err))
}

pub fn handle_websocket_close_log_err(socket: &WebSocketId, engine: &mut LuaState) {
    match handle_websocket_close(socket, engine) {
        Ok(()) => {}
        Err(e) => console::log_2(
            &JsValue::from_str("An error occurred calling the WebSocket input handler"),
            &JsValue::from_str(&e),
        ),
    }
}

pub(super) fn sendmud_raw<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    _global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let mud: WebSocketId = try_unwrap_socketid(ctx, &stack.pop_front())?;
        let msg: piccolo::String = stack.from_front(ctx)?;
        send_message_to_mud(&mud, msg.as_bytes(), &global_memo)?;
        Ok(piccolo::CallbackReturn::Return)
    })
}

pub(super) fn connect_mud<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    _global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let mut trusted: bool = false;
        let name: String = loop {
            let v: String = stack.from_front(ctx)?;
            if v == "-trust" {
                trusted = true;
                continue;
            }
            break v;
        };
        let name: Value<'gc> = ctx.intern(debrace(&name).as_bytes()).into();

        let muds = Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"muds")))?;
        if !muds.get_value(name).is_nil() {
            Err(Error::msg(
                "Attempt to create MUD connection using name that's already taken",
            ))?
        }

        let url: String = stack.from_front(ctx)?;
        let new_socket = intern_id(
            ctx,
            connect_websocket(
                trusted,
                &url,
                &mut global_memo.ws_registry.borrow_mut(),
                &global_memo,
            )?,
        );

        muds.set(ctx, name, new_socket)?;

        let conntab = Table::new(&ctx);
        muds.set(ctx, new_socket, conntab)?;

        Ok(piccolo::CallbackReturn::Return)
    })
}

pub(super) fn close_mud<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    _global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let mud_value: Value<'gc> = stack.pop_front();
        let socket_id = try_unwrap_socketid(ctx, &mud_value)?;
        match global_memo.ws_registry.borrow_mut().get_mut(&socket_id) {
            None => Err(Error::msg("That MUD connection doesn't exist."))?,
            Some(v) => {
                if v.closed {
                    Err(Error::msg("That MUD connection was already closed."))?
                }
                v.connection.close().unwrap_or(());
                v.retained_closures = None;
                v.closed = true;
            }
        }
        Ok(piccolo::CallbackReturn::Return)
    })
}

pub(super) fn delete_mud<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    _global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let name: String = stack.from_front(ctx)?;
        let name: Value<'gc> = ctx.intern(debrace(&name).as_bytes()).into();

        let muds = Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"muds")))?;
        let mud_value = muds.get_value(name);
        if mud_value.is_nil() {
            Err(Error::msg(
                "Attempt to delete MUD connection that wasn't found",
            ))?
        }

        let socket_id = try_unwrap_socketid(ctx, &mud_value)?;

        // Delete the MUD data if possible.
        let _ = muds.set_value(&ctx, mud_value, Value::Nil);
        for (k, v) in muds.iter() {
            match UserData::from_value(ctx, v) {
                Err(_) => continue,
                Ok(ud) => match ud.downcast::<Rootable!['gcb => Gc<'gcb, WebSocketId>]>() {
                    Err(_) => continue,
                    Ok(v) if v.as_ref() != &socket_id => continue,
                    _ => {}
                },
            }
            let _ = muds.set_value(&ctx, k, Value::Nil);
        }
        unintern_id(ctx, &socket_id);
        match global_memo.ws_registry.borrow_mut().remove(&socket_id) {
            None => {}
            Some(sockdat) if sockdat.closed => {}
            Some(sockdat) => sockdat.connection.close().unwrap_or(()),
        }

        Ok(piccolo::CallbackReturn::Return)
    })
}

pub(super) fn mudoutput<'gc>(ctx: Context<'gc>, global_memo: &GlobalMemoCell) -> Callback<'gc> {
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        // Temporary hack for testing... alias to echo
        let echo = Function::from_value(
            ctx,
            Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"commands")))?
                .get(ctx, ctx.intern_static(b"echo")),
        )?;

        let mud: Value<'gc> = stack.pop_front();
        let output: piccolo::String<'gc> = stack.from_front(ctx)?;

        let conntab =
            Table::from_value(ctx, ctx.get_global(ctx.intern_static(b"muds")))?.get_value(mud);
        if conntab.is_nil() {
            Err(Error::msg("Received input from MUD not in muds table."))?
        }

        Ok(piccolo::CallbackReturn::Call {
            function: echo,
            then: None,
        })
    })
}
